/*

        CONVERTS A FIXED LENGTH IMPORT ASCII FILE TO AN ASCII DELIMITED CSV FILE
                             JEAN-CLAUDE BERGER, OCTOBER 1993
                                      VERSION 1.0

           Compiled with Microsoft Visual C++ for NT 1.10 or Microsoft C/C++ 7
                             Uses Microsoft Foundation Classes


 Legal  : Non commercial users :

             Feel free to use this program as you want as long as you don't sell it
             or include it in a commercial package and don't remove the author's name. 
             If you like it, send me a postcard with a photography of your city. 
             Here is my address :

                Jean-Claude Berger
                6, rue Moliere
                69006 LYON 
                FRANCE

          Commercial, governmental or other institutions who wish to make use of 
          this program should contact the author for licensing details.

          This program is supplied "as is" with no warranty of any kind. The author has 
          done his best to make sure the program is bug free and works as intended, but 
          if it trashes your hard disk, that's your problem.

 Usage  : FIX2DEL sourceFile paramFile destFile numberOfLinesToSkip

          SourceFile is the name of an ASCII file with fixed lenth fields. For
          example a file containing :

              999MARTIN    FRANCE   10.33
             1000DOE       USA       9.5

          ParamFile is the name of an ASCII file containing the lenth of each field
          separated by a CR/LF. For example :

             7
             10
             6
             8

             You may type blank lines, they will be ignored or you may begin a line
             with a semi-colon (;) for writing comments. The comment should not
             exceed 40 characters per line. For example :

             ; CustId
             7

             ; Name
             10
             ; Country
             6
             8

          DestFile is the file created by the conversion. It would be in this form:

             "999","MARTIN","FRANCE","10.3"
             "1000","DOE","USA",9.5"

             WARNING : IF A FILE WITH THE SAME NAME EXISTS, IT WILL BE OVERWRITTEN.


          NumberOfLinesToSkip is the number of lines at the begining of the source file 
          that will be ignored by the program. So the source file may have an header. 
          For example :

             OCTOBER, 1993 FROM THE MARKETING DEPARTMENT

                 999MARTIN    FRANCE   10.33
                1000DOE       USA       9.5

             The NumberOfLinesToSkip would be 2 in that case.

 Limits : The ASCII files can't have more than 255 fields and the lenth of one
          line of the source file can't exceed 9999 characters (see the macro TAILLELIGNE).
          A comment in the parameters file should not exceed 40 characters.

*/

#ifndef _WIN32
   #define _DOS
#endif


#include "afx.h"
#include "stdlib.h"
#include "iostream.h"
#include "conio.h"
#include "string.h"

#define TAILLELIGNE 9999              // Max size of a line in the source file
#define TAILLEPARAM 80

#ifdef _WIN32
   #include "windef.h"
#else
   #define TRUE 1
   #define FALSE 0
   #define BOOL int
#endif


unsigned int glarLongueurs[255];   // Receive the fixed fields lengths from the parameters file
int          nombreLongueurs = 0;  // Number of fields read from the parameters file

void quit(int codeErreur);


// ------------------------------------------------------------------------------------------
void main(int argc, char *argv[]) {

   int  traiteFichier(CStdioFile *pfOrigine, CStdioFile *pfParam, CStdioFile *pfDest,int linesToSkip);
   int  codeErreur = 2;

   if(argc < 5) quit(1);

   TRY   
   {
      CStdioFile  fOrigine(argv[1], CFile::modeRead | CFile::typeText);
      codeErreur++;
      CStdioFile  fParam(argv[2], CFile::modeRead | CFile::typeText);
      codeErreur++;
      CStdioFile  fDest(argv[3], CFile::modeCreate | CFile::typeText);
      codeErreur++;
      
      quit(traiteFichier(&fOrigine, &fParam, &fDest, atoi(argv[4])));
   }
      
   CATCH (CFileException, e) 
   {
      quit(codeErreur);
   }
   END_CATCH

}

// ------------------------------------------------------------------------------------------

int traiteFichier(CStdioFile *pfOrigine, CStdioFile *pfParam, CStdioFile *pfDest, int linesToSkip) 
{
   // Treats the data file if we can read the parameter file.

   BOOL lireParametres(CStdioFile *pfParam);
   int convertit(CStdioFile *pfOrigine, CStdioFile *pfDest, int linesToSkip);

   // The first thing to do is to read the parameter file to create an array with
   // the field lengthes

   if(lireParametres(pfParam)) 
   {
      // At this point, the array glarLongueurs holds the records lengths
      return convertit(pfOrigine, pfDest, linesToSkip);         
   }
   else return 4;


}
// ------------------------------------------------------------------------------------------
int convertit(CStdioFile *pfOrigine, CStdioFile *pfDest, int linesToSkip)
{

   // Treats the data file using the global array of parameters containing the lenghts

   unsigned char *pBufOrg, *pBufDest;
   int traiteChaine(unsigned char *pBufOrg, unsigned char *pBufDest, CStdioFile *pfDest);

   int i, 
       retVal = 0;


                 
   TRY 
   {
      // We need an input buffer and another space to copy the converted string

      if((pBufOrg = (unsigned char *) malloc(sizeof(unsigned char) * TAILLELIGNE)) != NULL)   
      {
         if((pBufDest = (unsigned char *) malloc(sizeof(unsigned char) * TAILLELIGNE + 256 * 3)) != NULL) 
         {
            // We skip the x lines of the input file (as stated in the last command line parameter)
            for(i = 0; i < linesToSkip ; i++) 
            {
               pfOrigine->ReadString((char *)pBufOrg, TAILLELIGNE - 1);
            } 

            // Now, we read the source file until its end
            while(pfOrigine->ReadString((char *)pBufOrg, TAILLELIGNE - 1) != NULL)
            {
               if((retVal = traiteChaine(pBufOrg, pBufDest, pfDest)))
               {
                  break; // If there is an error during the string conversion
               }
            }
            pfOrigine->Close();
            pfDest->Close();
         }
         else retVal = 5;
      }
      else retVal = 5;
   }

   CATCH (CFileException, e) 
   {
      retVal = 7;

   } END_CATCH
   
   // I must free memory here and not inside the if statement concerning the malloc
   // because if there's an exception, I may not execute the end of the if block.

   if (pBufDest != NULL) free(pBufDest);
   if (pBufOrg  != NULL) free(pBufOrg);

   
   return retVal;
                   
}


// ------------------------------------------------------------------------------------------

int traiteChaine(unsigned char *pBufOrg, unsigned char *pBufDest, CStdioFile *pfDest)
{

   BOOL isBlankDate(unsigned char *pDebutChamp, unsigned char *pFinChamp);
   register unsigned char *pFinBufDest, *pDebutChamp, *pFinChamp, *pDebutBufDest;

   register int i;
   int debutChamp = 0;

   pDebutBufDest = pBufDest;
   pFinBufDest = pBufDest + TAILLELIGNE + 256 * 3;

   // We skip any blank line. They have a one byte length (because of the CR/LF)
   #ifdef _WIN32
      if(strlen((char *)pBufOrg) < 2 ) return 0;
   #else
      if(_fstrlen((char  *)pBufOrg) < 2) return 0;
   #endif      


   for( i = 0; i < nombreLongueurs ; i++ )
   {
      pDebutChamp = pBufOrg + debutChamp; //Points on the first char of the field
      pFinChamp   = pDebutChamp + glarLongueurs[i] - 1; // Points on the last one

      // Looks for any non space character
      while(*pDebutChamp == ' ' && pDebutChamp < pFinChamp) pDebutChamp++;
      while(*pFinChamp == ' ' && pFinChamp >= pDebutChamp) pFinChamp--;

      // pDebutChamp points on the first non space character and pFinChamp on the last one
      // unless there are only spaces. In this case, pFinChamp is SMALLER that pDebutChamp
      
      // First, we write a comma if this not the first field then the double quote to
      // the destination buffer
      if((i > 0) && (pBufDest < pFinBufDest)) *pBufDest++ = ',';
      if(pBufDest < pFinBufDest) *pBufDest++ = '"';
      if (pBufDest >= pFinBufDest) return 9;


      /*  1993.11.22 There is a special case : Some source files have blank dates 
          in the form "/  /". They should be translated in "". So we examine the 
          string pointed by pDebutChamp / pFinChamp. If it is "/  /" then we
          do not copy the string to the destination buffer. We do not need to 
          modify pDebutChamp / pFinChamp themselves as they are not used later.
       */

      if(!isBlankDate(pDebutChamp, pFinChamp)) 
      {
          // Now we copy the string itself
         while(pDebutChamp <= pFinChamp)
         {
            *pBufDest++ = *pDebutChamp++;
            if (pBufDest >= pFinBufDest) return 9;
         }
      }

      // Now, we have to write the closing double quote
      if(pBufDest < pFinBufDest) *pBufDest++ = '"';
      if (pBufDest >= pFinBufDest) return 9;

      debutChamp += glarLongueurs[i];
   } // endFor
   
   // After the last field, we have to write a CR and end the string with 0 char 
   // then write it to the file
   if(pBufDest < pFinBufDest) *pBufDest++ = '\n';
   if(pBufDest < pFinBufDest) *pBufDest++ = '\0';
   if (pBufDest >= pFinBufDest) return 9;

   TRY
   {
      pfDest->WriteString((char *) pDebutBufDest);
   }

   CATCH (CFileException, e) 
   {
      return 11;
   }
   END_CATCH

   return 0;
}

// ------------------------------------------------------------------------------------------

BOOL lireParametres(CStdioFile *pfParam) {

   // We load the lengths in the global array glarLongueurs


   char bufferNombre[TAILLEPARAM];          //  buffer for reading comments or lenghts
// global int nombreLongueurs = 0; 
// global unsigned int glarLongueurs[255];

   TRY
   {
      while((pfParam->ReadString(bufferNombre, TAILLEPARAM)!= NULL) && (nombreLongueurs < 255))
      {
         // Skipping blank lines or comment lines (line begining with ;)

         if ((strlen(bufferNombre) > 1) && (*bufferNombre != ';')) 
         {                          
            glarLongueurs[nombreLongueurs++] = atoi(bufferNombre);
         }
      }
      pfParam->Close();
      return TRUE; 
   } 


   CATCH (CFileException, e) {
   }
   END_CATCH

   return FALSE;  // Should be called only if there is an exception
}

// ------------------------------------------------------------------------------------------
void quit(int codeErreur) {

   _putch('\n');

   switch(codeErreur) {

      case 0:
         cout << "Conversion successfull";
         break;
      case 1:
         cerr << "Usage : FIX2DELI SourceFile ParamFile DestFile numberOfLinesToSkip\n";
         cerr << "\n                Copyright Jean-Claude Berger, 1993";
         cerr << "\n            Converts a fixed length fields file to CSV";
         break;
      case 2:
         cerr << "Error opening source file";
         break;
      case 3:
         cerr << "Error opening parameter file";
         break;
      case 4:
         cerr << "Error reading parameters file";
         break;
      case 5:
         cerr << "Memory allocation error";
         break;
      case 9:
         cerr << "Output buffer overflow";
         break;
      case 11:
         cerr << "Error writing destination file";
         break;
      default:
         cerr << "Unexpected error";
         break;
   } 
   _putch('\n');
   exit(codeErreur);
}

// ------------------------------------------------------------------------------------------
BOOL isBlankDate(unsigned char *pDebutChamp, unsigned char *pFinChamp) {
   
   // Determines if the char "string" pointed by pDebutChamp and pFinChamp is "/  /".
   // Note that it is not a real '\0' terminated string, so we can't use strcomp.

   BOOL retVal = TRUE;
   unsigned char *pDateVide, *pFinDateVide;


   pDateVide = (unsigned char *) "/  /";
   retVal = TRUE;

   #ifdef _WIN32
      pFinDateVide = (unsigned char *)pDateVide + strlen((char *) pDateVide);
   #else
      pFinDateVide = (unsigned char *)pDateVide + _fstrlen((char *) pDateVide);
   #endif      

   // If the two "strings" don't have the same length, they don't match of course.
   // (We substract 1 because dateVide is a \0 terminated string and Champ is not.)

   if( ( pFinChamp - pDebutChamp ) != ( ( (pFinDateVide - 1) - pDateVide ) ) ) 
   {
      return FALSE;
   }

   while(pDebutChamp <= pFinChamp) 
   {
      if((*pDebutChamp++ != *pDateVide++)) 
      {
         // There is a different character
         retVal = FALSE;
         break;
      }

   }

   return retVal;
}
